from functools import lru_cache
from typing import Callable, Iterable, Dict, Any

import numpy
from numpy import ndarray, mean

from pjdata.content.data import Data
from pjdata.transformer.pholder import PHolder
from pjdata.types import Result
from pjml.abs.component import Component
from pjml.abs.mixin.functioninspection import withFunctionInspection
from pjml.stream.reduce.accumulator import Accumulator


class InterruptedStreamException(Exception):
    pass


class Summ(Component, withFunctionInspection):
    """Given a field, summarizes a Collection object to a Data object.

    The resulting Data object will have only the 's' field. To keep other
    fields, consider using a Keep containing all the concurrent part:
    Keep(Expand -> ... -> Summ).

    The cells of the given field (matrix) will be averaged across all data
    objects, resulting in a new matrix with the same dimensions.
    """

    def __init__(self, field: str = "R", function: str = "mean", **kwargs):
        config = self._to_config(locals())
        super().__init__(config, deterministic=True, **kwargs)
        self.function = Summ.function_from_name()[config["function"]]
        self.field = field

    @lru_cache()
    def _enhancer_info(self, data: Data = None) -> Dict[str, Any]:
        return {}

    @lru_cache()
    def _model_info(self, data: Data) -> Dict[str, Any]:
        return {}

    def _enhancer_func(self) -> Callable[[Data], Result]:
        summarize = self.function

        def transform(data: Data) -> Result:
            def step(d, acc):
                if d.isfrozen or d.failure:
                    return d.transformedby(PHolder(self)), None
                acc.append(d.field(self.field, "Summ"))
                return d, acc

            iterator = Accumulator(data.stream, start=[], step_func=step, summ_func=summarize)

            def lazy():
                # try:
                return iterator.result

            # except StreamException:

            return {"stream": iterator, "S": lazy}

        return transform

    def _model_func(self, data: Data) -> Callable[[Data], Result]:
        return self._enhancer_func()

    @staticmethod
    def _fun_mean(values: Iterable[float]) -> ndarray:
        res = mean([m for m in values], axis=0)
        return numpy.array(res) if isinstance(res, tuple) else res
